/*
 * Copyright 1994, 1995 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <Unidraw/catalog.h>
#include <Unidraw/unidraw.h>
#include <Unidraw/Graphic/graphic.h>

#include <Vpf/vpffclass.h>
#include <Vpf/vpffeature.h>
#include <VpfUnidraw/vpfviewstate.h>

#include <Attribute/attribute.h>
#include <Attribute/attrlist.h>
#include <Attribute/attrvalue.h>

#include <InterViews/transformer.h>

#include <limits.h>

/****************************************************************************/

Graphic* VpfViewState::_poocean_gs = nil;
Graphic* VpfViewState::_poland_gs = nil;
Graphic* VpfViewState::_polbnd_gs = nil;
Graphic* VpfViewState::_poice_gs = nil;
Graphic* VpfViewState::_poline_gs = nil;
Graphic* VpfViewState::_pppoint_gs = nil;
Graphic* VpfViewState::_dnarea_gs = nil;
Graphic* VpfViewState::_dnline_gs = nil;
Graphic* VpfViewState::_dnpoint_gs = nil;
Graphic* VpfViewState::_dnnull_gs = nil;
Graphic* VpfViewState::_hyline_gs = nil;
Graphic* VpfViewState::_tileref_gs = nil;
Graphic* VpfViewState::_aepoint_gs = nil;
Graphic* VpfViewState::_clarea_gs = nil;
Graphic* VpfViewState::_clline_gs = nil;
Graphic* VpfViewState::_clpoint_gs = nil;
Graphic* VpfViewState::_dslake_gs = nil;
Graphic* VpfViewState::_dsland_gs = nil;
Graphic* VpfViewState::_hyarea1_gs = nil;
Graphic* VpfViewState::_hyarea2_gs = nil;
Graphic* VpfViewState::_hyarea3_gs = nil;
Graphic* VpfViewState::_hyarea4_gs = nil;
Graphic* VpfViewState::_hyarea5_gs = nil;
Graphic* VpfViewState::_hyarea6_gs = nil;
Graphic* VpfViewState::_hyarea_gs = nil;
Graphic* VpfViewState::_hypoint_gs = nil;
Graphic* VpfViewState::_hsline_gs = nil;
Graphic* VpfViewState::_hspoint_gs = nil;
Graphic* VpfViewState::_lcarea_gs = nil;
Graphic* VpfViewState::_lcpoint_gs = nil;
Graphic* VpfViewState::_ofline_gs = nil;
Graphic* VpfViewState::_ofpoint_gs = nil;
Graphic* VpfViewState::_phline_gs = nil;
Graphic* VpfViewState::_pparea_gs = nil;
Graphic* VpfViewState::_rrline_gs = nil;
Graphic* VpfViewState::_rdline_gs = nil;
Graphic* VpfViewState::_tsline_gs = nil;
Graphic* VpfViewState::_tspoint_gs = nil;
Graphic* VpfViewState::_utline_gs = nil;
Graphic* VpfViewState::_vgrange_gs = nil;
Graphic* VpfViewState::_vgforst_gs = nil;
Graphic* VpfViewState::_vgtundr_gs = nil;
Graphic* VpfViewState::_unkarea_gs = nil;
Graphic* VpfViewState::_unkline_gs = nil;
Graphic* VpfViewState::_unkpoint_gs = nil;
Graphic* VpfViewState::_text_gs = nil;
Graphic* VpfViewState::_none_gs = nil;
Graphic* VpfViewState::_exarea_gs = nil;
Graphic* VpfViewState::_expoint_gs = nil;
Graphic* VpfViewState::_inpoint_gs = nil;
Graphic* VpfViewState::_utpoint_gs = nil;
Graphic* VpfViewState::_stpoint_gs = nil;
Graphic* VpfViewState::_dangerpoint_gs = nil;
Graphic* VpfViewState::_elpoint_gs = nil;

VpfViewState::VpfViewState(PJ* p, double xmin, double ymin, double scale)
: MapViewState(p)
{
    UV data;
    data.u = xmin * DEG_TO_RAD;
    data.v = ymin * DEG_TO_RAD;
    data = pj_fwd(data, p);
    int x = round(data.u);
    int y = round(data.v);
    Transformer* trans = new Transformer(scale,
					 0.0,
					 0.0,
					 scale,
					 -x*scale,
					 -y*scale);
    settrans(trans);
}

VpfViewState::VpfViewState() 
: MapViewState()
{
}

VpfViewState::~VpfViewState() {
}

Transformer* VpfViewState::deftrans() const {
    return _trans;
}

Graphic* VpfViewState::themegs(MapFeature* mf) const {
    if (!_poocean_gs) {
	Catalog* catalog = unidraw->GetCatalog();

	_none_gs = new FullGraphic(stdgraphic);
	_none_gs->SetPattern(catalog->FindNonePattern());
	_none_gs->FillBg(false);
	_none_gs->SetBrush(catalog->FindNoneBrush());

	_poocean_gs = new FullGraphic(stdgraphic);
	_poocean_gs->SetPattern(catalog->FindGrayLevel(1.0));
	PSColor* ocean = catalog->FindColor("LightSkyBlue");
	PSColor* black = catalog->FindColor("black");
	_poocean_gs->SetColors(black,ocean);

	_poland_gs = new FullGraphic(stdgraphic);
	_poland_gs->SetPattern(catalog->FindGrayLevel(1.0));
	PSColor* land = catalog->FindColor("PeachPuff");
	_poland_gs->SetColors(black,land);

	_polbnd_gs = new FullGraphic(stdgraphic);
	_polbnd_gs->SetPattern(catalog->FindGrayLevel(1.0));
	PSColor* landb = catalog->FindColor("PeachPuff");
	_polbnd_gs->SetColors(black,landb);

	_poice_gs = new FullGraphic(stdgraphic);
	PSColor* white = catalog->FindColor("white");
	_poice_gs->SetPattern(catalog->FindGrayLevel(0.1));
	_poice_gs->SetColors(black,white);

	_poline_gs = new FullGraphic(stdgraphic);
	_poline_gs->SetPattern(catalog->FindNonePattern());
	_poline_gs->SetColors(black,black);

	_pppoint_gs = new FullGraphic(stdgraphic);
	PSColor* gray = catalog->FindColor("gray66");
	_pppoint_gs->SetColors(black,gray);
	_pppoint_gs->SetPattern(catalog->FindGrayLevel(1.0));

	_dnarea_gs = new FullGraphic(stdgraphic);
	PSColor* lightblue = catalog->FindColor("LightSkyBlue");
	PSColor* royalblue = catalog->FindColor("RoyalBlue");
	_dnarea_gs->SetColors(royalblue, lightblue);
	_dnarea_gs->SetPattern(catalog->FindGrayLevel(1.0));

	_dnline_gs = new FullGraphic(stdgraphic);
	_dnline_gs->SetColors(royalblue, royalblue);
	_dnline_gs->SetPattern(catalog->FindNonePattern());

	_dnpoint_gs = new FullGraphic(stdgraphic);
	_dnpoint_gs->SetColors(royalblue, royalblue);
	_dnpoint_gs->SetPattern(catalog->FindGrayLevel(1.0));

	_dnnull_gs = new FullGraphic(stdgraphic);
	_dnnull_gs->SetColors(royalblue, white);
	_dnnull_gs->SetPattern(catalog->FindGrayLevel(0.9));
	_dnnull_gs->FillBg(false);

	_hyline_gs = new FullGraphic(stdgraphic);
	PSColor* khaki = catalog->FindColor("chocolate");
	_hyline_gs->SetColors(khaki, khaki);
	_hyline_gs->SetPattern(catalog->FindNonePattern());

	_tileref_gs = new FullGraphic(stdgraphic);
	_tileref_gs->SetColors(black,black);
	_tileref_gs->SetPattern(catalog->ReadPattern("pattern",7));
	_tileref_gs->FillBg(false);

	_aepoint_gs = new FullGraphic(stdgraphic);
	PSColor* nblue = catalog->FindColor("NavyBlue");
	_aepoint_gs->SetColors(black, nblue);
	_aepoint_gs->SetPattern(catalog->FindGrayLevel(1.0));

	_clarea_gs = new FullGraphic(stdgraphic);
	PSColor* sgray = catalog->FindColor("SlateGray");
	_clarea_gs->SetColors(sgray, white);
	_clarea_gs->SetPattern(catalog->FindGrayLevel(0.7));
	_clarea_gs->FillBg(false);

	_clline_gs = new FullGraphic(stdgraphic);
	_clline_gs->SetColors(sgray, sgray);
	_clline_gs->SetPattern(catalog->FindNonePattern());

	_clpoint_gs = new FullGraphic(stdgraphic);
	_clpoint_gs->SetColors(black, sgray);
	_clpoint_gs->SetPattern(catalog->FindGrayLevel(1.0));

	_dslake_gs = new FullGraphic(stdgraphic);
	_dslake_gs->SetColors(royalblue, lightblue);
	_dslake_gs->SetPattern(catalog->FindGrayLevel(1.0));

	_dsland_gs = new FullGraphic(stdgraphic);
	_dsland_gs->SetColors(black, land);
	_dsland_gs->SetPattern(catalog->FindGrayLevel(1.0));

	PSColor* sdlbrn = catalog->FindColor("SaddleBrown");
	PSColor* dkgldrd = catalog->FindColor("DarkGoldenrod");
	PSColor* gldrd = catalog->FindColor("goldenrod");
	PSColor* gold = catalog->FindColor("gold");
	PSColor* ltgldrd = catalog->FindColor("PaleGoldenrod");
	PSColor* plgldrd = catalog->FindColor("LightGoldenrodYellow");

	_hyarea1_gs = new FullGraphic(stdgraphic);
	_hyarea1_gs->SetColors(sdlbrn, sdlbrn);
	_hyarea1_gs->FillBg(false);
	_hyarea1_gs->SetPattern(catalog->FindGrayLevel(0.5));

	_hyarea2_gs = new FullGraphic(stdgraphic);
	_hyarea2_gs->SetColors(dkgldrd, dkgldrd);
	_hyarea2_gs->FillBg(false);
	_hyarea2_gs->SetPattern(catalog->FindGrayLevel(0.5));

	_hyarea3_gs = new FullGraphic(stdgraphic);
	_hyarea3_gs->SetColors(gldrd, gldrd);
	_hyarea3_gs->FillBg(false);
	_hyarea3_gs->SetPattern(catalog->FindGrayLevel(0.5));

	_hyarea4_gs = new FullGraphic(stdgraphic);
	_hyarea4_gs->SetColors(gold, gold);
	_hyarea4_gs->FillBg(false);
	_hyarea4_gs->SetPattern(catalog->FindGrayLevel(0.5));

	_hyarea5_gs = new FullGraphic(stdgraphic);
	_hyarea5_gs->SetColors(ltgldrd, ltgldrd);
	_hyarea5_gs->FillBg(false);
	_hyarea5_gs->SetPattern(catalog->FindGrayLevel(0.5));

	_hyarea6_gs = new FullGraphic(stdgraphic);
	_hyarea6_gs->SetColors(plgldrd, plgldrd);
	_hyarea6_gs->FillBg(false);
	_hyarea6_gs->SetPattern(catalog->FindGrayLevel(0.5));

	_hyarea_gs = new FullGraphic(stdgraphic);
	_hyarea_gs->SetColors(black, black);
	_hyarea_gs->FillBg(false);
	_hyarea_gs->SetPattern(catalog->ReadPattern("pattern",8));

	_hypoint_gs = new FullGraphic(stdgraphic);
	_hypoint_gs->SetColors(black, khaki);
	_hypoint_gs->SetPattern(catalog->FindGrayLevel(1.0));

	_hsline_gs = new FullGraphic(stdgraphic);
	PSColor* sienna = catalog->FindColor("sienna");
	_hsline_gs->SetColors(sienna, sienna);
	_hsline_gs->SetPattern(catalog->FindNonePattern());

	_hspoint_gs = new FullGraphic(stdgraphic);
	_hspoint_gs->SetColors(white, khaki);
	_hspoint_gs->SetPattern(catalog->FindGrayLevel(1.0));

	_lcarea_gs = new FullGraphic(stdgraphic);
	PSColor* cblue = catalog->FindColor("CadetBlue");
	_lcarea_gs->SetColors(cblue, cblue);
	_lcarea_gs->SetPattern(catalog->FindGrayLevel(0.9));
	_lcarea_gs->FillBg(false);

	_lcpoint_gs = new FullGraphic(stdgraphic);
	_lcpoint_gs->SetColors(black, cblue);
	_lcpoint_gs->SetPattern(catalog->FindGrayLevel(1.0));

	_ofline_gs = new FullGraphic(stdgraphic);
	PSColor* ired = catalog->FindColor("IndianRed");
	_ofline_gs->SetColors(ired, ired);
	_ofline_gs->SetPattern(catalog->FindNonePattern());

	_ofpoint_gs = new FullGraphic(stdgraphic);
	_ofpoint_gs->SetColors(black, ired);
	_ofpoint_gs->SetPattern(catalog->FindGrayLevel(1.0));

	_phline_gs = new FullGraphic(stdgraphic);
	PSColor* dblue = catalog->FindColor("DarkSlateBlue");
	_phline_gs->SetColors(dblue, dblue);
	_phline_gs->SetPattern(catalog->FindNonePattern());

	_pparea_gs = new FullGraphic(stdgraphic);
	_pparea_gs->SetColors(black, gray);
	_pparea_gs->SetPattern(catalog->FindGrayLevel(0.75));
	_pparea_gs->FillBg(false);

	_rrline_gs = new FullGraphic(stdgraphic);
	_rrline_gs->SetColors(gray, gray);
	_rrline_gs->SetPattern(catalog->FindNonePattern());
	_rrline_gs->SetBrush(catalog->ReadBrush("brush", 8));
	_rrline_gs->FillBg(false);

	_rdline_gs = new FullGraphic(stdgraphic);
	PSColor* tred = catalog->FindColor("tomato");
	_rdline_gs->SetColors(tred, tred);
	_rdline_gs->SetPattern(catalog->FindNonePattern());

	_tsline_gs = new FullGraphic(stdgraphic);
	PSColor* mar = catalog->FindColor("maroon");
	_tsline_gs->SetColors(mar, mar);
	_tsline_gs->SetPattern(catalog->FindNonePattern());

	_tspoint_gs = new FullGraphic(stdgraphic);
	_tspoint_gs->SetColors(black, mar);
	_tspoint_gs->SetPattern(catalog->FindGrayLevel(1.0));

	_utline_gs = new FullGraphic(stdgraphic);
	PSColor* dr = catalog->FindColor("DarkOrchid");
	_utline_gs->SetColors(dr, dr);
	_utline_gs->SetPattern(catalog->FindNonePattern());

	_vgforst_gs = new FullGraphic(stdgraphic);
	PSColor* gr = catalog->FindColor("ForestGreen");
	_vgforst_gs->SetColors(gr,gr);
	_vgforst_gs->SetPattern(catalog->FindGrayLevel(0.3));
	_vgforst_gs->FillBg(false);

	_vgrange_gs = new FullGraphic(stdgraphic);
	_vgrange_gs->SetColors(gr,gr);
	_vgrange_gs->SetPattern(catalog->FindGrayLevel(0.6));
	_vgrange_gs->FillBg(false);

	_vgtundr_gs = new FullGraphic(stdgraphic);
	_vgtundr_gs->SetColors(gr,gr);
	_vgtundr_gs->SetPattern(catalog->FindGrayLevel(0.9));
	_vgtundr_gs->FillBg(false);

	_unkarea_gs = new FullGraphic(stdgraphic);
	_unkarea_gs->SetColors(black,black);
	_unkarea_gs->SetPattern(catalog->FindGrayLevel(0.9));
	_unkarea_gs->FillBg(false);

	_unkline_gs = new FullGraphic(stdgraphic);
	_unkline_gs->SetPattern(catalog->FindNonePattern());

	_unkpoint_gs = new FullGraphic(stdgraphic);

	_text_gs = new FullGraphic(stdgraphic);
	_text_gs->SetColors(black, white);
	_text_gs->FillBg(false);

	_exarea_gs = new FullGraphic(stdgraphic);
	_exarea_gs->SetColors(ired, ired);
	_exarea_gs->FillBg(false);
	_exarea_gs->SetPattern(catalog->FindGrayLevel(0.5));

	_expoint_gs = new FullGraphic(stdgraphic);
	_expoint_gs->SetColors(black, ired);
	_expoint_gs->SetPattern(catalog->FindGrayLevel(1.0));

	_inpoint_gs = new FullGraphic(stdgraphic);
	PSColor* org = catalog->FindColor("DarkOrange");
	_inpoint_gs->SetColors(black, org);
	_inpoint_gs->SetPattern(catalog->FindGrayLevel(1.0));

	_utpoint_gs = new FullGraphic(stdgraphic);
	_utpoint_gs->SetColors(black, dr);
	_utpoint_gs->SetPattern(catalog->FindGrayLevel(1.0));

	_stpoint_gs = new FullGraphic(stdgraphic);
	PSColor* pvr = catalog->FindColor("PaleVioletRed");
	_stpoint_gs->SetColors(black, pvr);
	_stpoint_gs->SetPattern(catalog->FindGrayLevel(1.0));

	_dangerpoint_gs = new FullGraphic(stdgraphic);
	PSColor* dp = catalog->FindColor("DeepPink");
	_dangerpoint_gs->SetColors(black, dp);
	_dangerpoint_gs->SetPattern(catalog->FindGrayLevel(1.0));

	_elpoint_gs = new FullGraphic(stdgraphic);
	_elpoint_gs->SetColors(black, gldrd);
	_elpoint_gs->SetPattern(catalog->FindGrayLevel(1.0));

    }

    VpfFeature* vf = (VpfFeature*)mf;
    VpfFeatureClass* vfc = vf->feature_class();
    if (vf->type() == MapFeature::TextType) {
	return _text_gs;
    }
    if (strcmp(vfc->name(), "graticule")==0)
	return _unkline_gs;
    if (strcmp(vfc->name(),"tileref")==0)
	return _none_gs;
    if (strcmp(vfc->name(),"poarea")==0 && vf->long_attr("POPYTYPE")==2) {
	return _poocean_gs;
    }
    else if (strcmp(vfc->name(),"poarea")==0 && vf->long_attr("POPYTYPE")==1) {
	return _poland_gs;
    }
    else if (strcmp(vfc->name(),"polbnd")==0) {
	return _polbnd_gs;
    }
    else if (strcmp(vfc->name(),"poarea")==0 &&
	     (vf->long_attr("POPYTYPE")==3 ||
	      vf->long_attr("POPYTYPE")==4 ||
	      vf->long_attr("POPYTYPE")==5)) {
	return _poice_gs;
    }
    else if (strcmp(vfc->name(),"poline")==0) {
	return _poline_gs;
    }
    else if (strcmp(vfc->name(),"pppoint")==0) {
	return _pppoint_gs;
    }
    else if (strcmp(vfc->name(),"dnarea")==0 &&
	     vf->long_attr("DNPYTYPE")==MISSING_VALUE) {
	return _dnnull_gs;
    }
    else if (strcmp(vfc->name(),"dnarea")==0) {
	return _dnarea_gs;
    }
    else if (strcmp(vfc->name(),"dnline")==0) {
	return _dnline_gs;
    }
    else if (strcmp(vfc->name(),"dnpoint")==0) {
	return _dnpoint_gs;
    }
    else if (strcmp(vfc->name(),"hyline")==0) {
	return _hyline_gs;
    }
    else if (strcmp(vfc->name(),"tileref")==0) {
	return _tileref_gs;
    }
    else if (strcmp(vfc->name(),"aepoint")==0) {
	return _aepoint_gs;
    }
    else if (strcmp(vfc->name(),"clarea")==0) {
	return _clarea_gs;
    }
    else if (strcmp(vfc->name(),"clline")==0) {
	return _clline_gs;
    }
    else if (strcmp(vfc->name(),"clpoint")==0) {
	return _clpoint_gs;
    }
    else if (strcmp(vfc->name(),"dspoint")==0 && vf->long_attr("DSPTTYPE")==1) {
	return _dslake_gs;
    }
    else if (strcmp(vfc->name(),"dspoint")==0 && vf->long_attr("DSPTTYPE")==2) {
	return _dsland_gs;
    }
    else if (strcmp(vfc->name(),"hyarea")==0) {
	switch (vf->long_attr("HYPYTYPE")) {
	case 1: return _hyarea1_gs;
	case 2: return _hyarea2_gs;
	case 3: return _hyarea3_gs;
	case 4: return _hyarea4_gs;
	case 5: return _hyarea5_gs;
	case 6: return _hyarea6_gs;
	default: return _hyarea_gs;
	}
    }
    else if (strcmp(vfc->name(),"hypoint")==0) {
	return _hypoint_gs;
    }
    else if (strcmp(vfc->name(),"hsline")==0) {
	return _hsline_gs;
    }
    else if (strcmp(vfc->name(),"hspoint")==0) {
	return _hspoint_gs;
    }
    else if (strcmp(vfc->name(),"lcarea")==0) {
	return _lcarea_gs;
    }
    else if (strcmp(vfc->name(),"lcpoint")==0) {
	return _lcpoint_gs;
    }
    else if (strcmp(vfc->name(),"ofline")==0) {
	return _ofline_gs;
    }
    else if (strcmp(vfc->name(),"ofpoint")==0) {
	return _ofpoint_gs;
    }
    else if (strcmp(vfc->name(),"phline")==0) {
	return _phline_gs;
    }
    else if (strcmp(vfc->name(),"pparea")==0) {
	return _pparea_gs;
    }
    else if (strcmp(vfc->name(),"rrline")==0) {
	return _rrline_gs;
    }
    else if (strcmp(vfc->name(),"rdline")==0) {
	return _rdline_gs;
    }
    else if (strcmp(vfc->name(),"tsline")==0) {
	return _tsline_gs;
    }
    else if (strcmp(vfc->name(),"tspoint")==0) {
	return _tspoint_gs;
    }
    else if (strcmp(vfc->name(),"utline")==0) {
	return _utline_gs;
    }
    else if (strcmp(vfc->name(),"vgarea")==0) {
	switch(vf->long_attr("VGPYTYPE")) {
	case 31:
	case 32:
	case 33: return _vgrange_gs;
	case 41:
	case 42:
	case 43: return _vgforst_gs;
	case 83: return _vgtundr_gs;
	default: return _unkarea_gs;
	}
    }

    // vmap0
    // new gses to do: (current default will be black/patterned areas)
    // barrier lines
    // ice cliff
    // bluff/cliff lines
    // political (sub) boundary lines
    Attribute* attr = vf->attrlist() ? vf->attrlist()->GetAttr("f_code") : nil;
    const char* fcname = vfc->name();
    const char* fcode = attr ? attr->Value()->string_ptr() : nil;
    if (fcode) {
    if (strcmp(fcode,"AA010")==0 && strcmp(fcname,"extracta")==0) {
      // extracta.aft f_code AA010 Mine/Quarry 
      return _exarea_gs;
    }
    if (strcmp(fcode,"AA010")==0 && strcmp(fcname,"extractp")==0) {
      // extractp.pft f_code AA010 Mine/Quarry 
      return _expoint_gs;
    }
    if (strcmp(fcode,"AA050")==0 && strcmp(fcname,"extractp")==0) {
      // extractp.pft f_code AA050 Well 
      return _expoint_gs;
    }
    if (strcmp(fcode,"AA052")==0 && strcmp(fcname,"extracta")==0) {
      // extracta.aft f_code AA052 Oil/Gas Field 
      return _exarea_gs;
    }
    if (strcmp(fcode,"AA052")==0 && strcmp(fcname,"extractp")==0) {
      // extractp.pft f_code AA052 Oil/Gas Field 
      return _expoint_gs;
    }
    if (strcmp(fcode,"AC000")==0 && strcmp(fcname,"misindp")==0) {
      // misindp.pft f_code AC000 Processing Plant/Treatment Plant 
      return _inpoint_gs;
    }
    if (strcmp(fcode,"AC040")==0 && strcmp(fcname,"misindp")==0) {
      // misindp.pft f_code AC040 Oil/Gas Facilities 
      return _inpoint_gs;
    }
    if (strcmp(fcode,"AD010")==0 && strcmp(fcname,"utilp")==0) {
      // utilp.pft f_code AD010 Power Plant 
      return _utpoint_gs;
    }
    if (strcmp(fcode,"AD030")==0 && strcmp(fcname,"utilp")==0) {
      // utilp.pft f_code AD030 Substation/Transformer Yard 
      return _utpoint_gs;
    }
    if (strcmp(fcode,"AI030")==0 && strcmp(fcname,"mispopp")==0) {
      // mispopp.pft f_code AI030 Camp 
      return _pppoint_gs;
    }
    if (strcmp(fcode,"AK160")==0 && strcmp(fcname,"mispopp")==0) {
      // mispopp.pft f_code AK160 Stadium/Amphitheatre 
      return _pppoint_gs;
    }
    if (strcmp(fcode,"AL015")==0 && strcmp(fcname,"mispopp")==0) {
      // mispopp.pft f_code AL015 Building 
      return _clarea_gs;
    }
    if (strcmp(fcode,"AL020")==0 && strcmp(fcname,"builtupa")==0) {
      // builtupa.aft f_code AL020 Built-Up Area 
      return _clarea_gs;
    }
    if (strcmp(fcode,"AL020")==0 && strcmp(fcname,"builtupp")==0) {
      // builtupp.pft f_code AL020 Built-Up Area 
      return _clarea_gs;
    }
    if (strcmp(fcode,"AL025")==0 && strcmp(fcname,"mispopp")==0) {
      // mispopp.pft f_code AL025 Cairn 
      return _clarea_gs;
    }
    if (strcmp(fcode,"AL105")==0 && strcmp(fcname,"mispopp")==0) {
      // mispopp.pft f_code AL105 Settlement 
      return _clarea_gs;
    }
    if (strcmp(fcode,"AL130")==0 && strcmp(fcname,"mispopp")==0) {
      // mispopp.pft f_code AL130 Monument 
      return _clarea_gs;
    }
    if (strcmp(fcode,"AL135")==0 && strcmp(fcname,"mispopa")==0) {
      // mispopa.aft f_code AL135 Native Settlement 
      return _clarea_gs;
    }
    if (strcmp(fcode,"AL135")==0 && strcmp(fcname,"mispopp")==0) {
      // mispopp.pft f_code AL135 Native Settlement 
      return _pppoint_gs;
    }
    if (strcmp(fcode,"AL200")==0 && strcmp(fcname,"mispopp")==0) {
      // mispopp.pft f_code AL200 Ruins 
      return _pppoint_gs;
    }
    if (strcmp(fcode,"AL240")==0 && strcmp(fcname,"misindp")==0) {
      // misindp.pft f_code AL240 Tower (Non-communication) 
      return _inpoint_gs;
    }
    if (strcmp(fcode,"AL260")==0 && strcmp(fcname,"barrierl")==0) {
      // barrierl.lft f_code AL260 Wall 

    }
    if (strcmp(fcode,"AL260")==0 && strcmp(fcname,"mispopp")==0) {
      // mispopp.pft f_code AL260 Wall 
      return _pppoint_gs;
    }
    if (strcmp(fcode,"AM010")==0 && strcmp(fcname,"storagep")==0) {
      // storagep.pft f_code AM010 Depot (Storage) 
      return _stpoint_gs;
    }
    if (strcmp(fcode,"AM070")==0 && strcmp(fcname,"storagep")==0) {
      // storagep.pft f_code AM070 Tank 
      return _stpoint_gs;
    }
    if (strcmp(fcode,"AM080")==0 && strcmp(fcname,"storagep")==0) {
      // storagep.pft f_code AM080 Water Tower 
      return _stpoint_gs;
    }
    if (strcmp(fcode,"AN010")==0 && strcmp(fcname,"dqline")==0) {
      // dqline.lft f_code AN010 Railroad 
      return _rrline_gs;
    }
    if (strcmp(fcode,"AN010")==0 && strcmp(fcname,"railrdl")==0) {
      // railrdl.lft f_code AN010 Railroad 
      return _rrline_gs;
    }
    if (strcmp(fcode,"AN010")==0 && strcmp(fcname,"railrdl")==0) {
      // railrdl.lft f_code AN010 Railroad 
      return _rrline_gs;
    }
    if (strcmp(fcode,"AN060")==0 && strcmp(fcname,"rryardp")==0) {
      // rryardp.pft f_code AN060 Railroad Yard/Marshalling Yard 
      return _rrline_gs;
    }
    if (strcmp(fcode,"AP030")==0 && strcmp(fcname,"dqline")==0) {
      // dqline.lft f_code AP030 Road 
      return _rdline_gs;
    }
    if (strcmp(fcode,"AP030")==0 && strcmp(fcname,"roadl")==0) {
      // roadl.lft f_code AP030 Road 
      return _rdline_gs;
    }
    if (strcmp(fcode,"AP030")==0 && strcmp(fcname,"roadl")==0) {
      // roadl.lft f_code AP030 Road 
      return _rdline_gs;
    }
    if (strcmp(fcode,"AP050")==0 && strcmp(fcname,"dqline")==0) {
      // dqline.lft f_code AP050 Trail 
      return _rdline_gs;
    }
    if (strcmp(fcode,"AP050")==0 && strcmp(fcname,"traill")==0) {
      // traill.lft f_code AP050 Trail 
      return _rdline_gs;
    }
    if (strcmp(fcode,"AQ010")==0 && strcmp(fcname,"mistranl")==0) {
      // mistranl.lft f_code AQ010 Aerial Cableway Lines/Ski Lift Line 
      return _tsline_gs;
    }
    if (strcmp(fcode,"AQ040")==0 && strcmp(fcname,"dqline")==0) {
      // dqline.lft f_code AQ040 Bridge/Overpass/Viaduct 
      return _rdline_gs;
    }
    if (strcmp(fcode,"AQ040")==0 && strcmp(fcname,"transtrc")==0) {
      // transtrc.pft f_code AQ040 Bridge/Overpass/Viaduct 
      return _tspoint_gs;
    }
    if (strcmp(fcode,"AQ040")==0 && strcmp(fcname,"transtrl")==0) {
      // transtrl.lft f_code AQ040 Bridge/Overpass/Viaduct 
      return _rdline_gs;
    }
    if (strcmp(fcode,"AQ064")==0 && strcmp(fcname,"transtrc")==0) {
      // transtrc.pft f_code AQ064 Causeway 
      return _tspoint_gs;
    }
    if (strcmp(fcode,"AQ064")==0 && strcmp(fcname,"transtrl")==0) {
      // transtrl.lft f_code AQ064 Causeway 
      return _rdline_gs;
    }
    if (strcmp(fcode,"AQ070")==0 && strcmp(fcname,"dqline")==0) {
      // dqline.lft f_code AQ070 Ferry Crossing 
      return _dnline_gs;
    }
    if (strcmp(fcode,"AQ070")==0 && strcmp(fcname,"transtrc")==0) {
      // transtrc.pft f_code AQ070 Ferry Crossing 
      return _tspoint_gs;
    }
    if (strcmp(fcode,"AQ070")==0 && strcmp(fcname,"transtrl")==0) {
      // transtrl.lft f_code AQ070 Ferry Crossing 
      return _dnline_gs;
    }
    if (strcmp(fcode,"AQ113")==0 && strcmp(fcname,"dqline")==0) {
      // dqline.lft f_code AQ113 Pipeline/Pipe 
      return _tsline_gs;
    }
    if (strcmp(fcode,"AQ113")==0 && strcmp(fcname,"pipel")==0) {
      // pipel.lft f_code AQ113 Pipeline/Pipe 
      return _tsline_gs;
    }
    if (strcmp(fcode,"AQ116")==0 && strcmp(fcname,"utilp")==0) {
      // utilp.pft f_code AQ116 Pumping Station 
      return _utpoint_gs;
    }
    if (strcmp(fcode,"AQ130")==0 && strcmp(fcname,"transtrc")==0) {
      // transtrc.pft f_code AQ130 Tunnel 
      return _tspoint_gs;
    }
    if (strcmp(fcode,"AQ130")==0 && strcmp(fcname,"transtrl")==0) {
      // transtrl.lft f_code AQ130 Tunnel 
      return _rdline_gs;
    }
    if (strcmp(fcode,"AT030")==0 && strcmp(fcname,"dqline")==0) {
      // dqline.lft f_code AT030 Power Transmission Line 
      return _utline_gs;
    }
    if (strcmp(fcode,"AT030")==0 && strcmp(fcname,"utill")==0) {
      // utill.lft f_code AT030 Power Transmission Line 
      return _utline_gs;
    }
    if (strcmp(fcode,"AT050")==0 && strcmp(fcname,"utilp")==0) {
      // utilp.pft f_code AT050 Communication Building 
      return _utpoint_gs;
    }
    if (strcmp(fcode,"AT060")==0 && strcmp(fcname,"dqline")==0) {
      // dqline.lft f_code AT060 Telephone Line/Telegraph Line 
      return _utline_gs;
    }
    if (strcmp(fcode,"AT060")==0 && strcmp(fcname,"utill")==0) {
      // utill.lft f_code AT060 Telephone Line/Telegraph Line 
      return _utline_gs;
    }
    if (strcmp(fcode,"AT080")==0 && strcmp(fcname,"utilp")==0) {
      // utilp.pft f_code AT080 Communication Tower 
      return _utpoint_gs;
    }
    if (strcmp(fcode,"BA010")==0 && strcmp(fcname,"coastl")==0) {
      // coastl.lft f_code BA010 Coastline/Shoreline 
      return _dnline_gs;
    }
    if (strcmp(fcode,"BA010")==0 && strcmp(fcname,"libref")==0) {
      // libref.lft f_code BA010 Coastline/Shoreline 
      return _dnline_gs;
    }
    if (strcmp(fcode,"BA030")==0 && strcmp(fcname,"miscp")==0) {
      // miscp.pft f_code BA030 Island 
      return _poland_gs;
    }
    if (strcmp(fcode,"BA040")==0 && strcmp(fcname,"oceansea")==0) {
      // oceansea.aft f_code BA040 Water (except Inland) 
      return _poocean_gs;
    }
    if (strcmp(fcode,"BB040")==0 && strcmp(fcname,"miscl")==0) {
      // miscl.lft f_code BB040 Breakwater/Groyne 
      return _dnline_gs;
    }
    if (strcmp(fcode,"BB190")==0 && strcmp(fcname,"mistranl")==0) {
      // mistranl.lft f_code BB190 Pier/Wharf/Quay 
      return _dnline_gs;
    }
    if (strcmp(fcode,"BB230")==0 && strcmp(fcname,"miscl")==0) {
      // miscl.lft f_code BB230 Seawall 
      return _dnline_gs;
    }
    if (strcmp(fcode,"BC050")==0 && strcmp(fcname,"mispopp")==0) {
      // mispopp.pft f_code BC050 Lighthouse 
      return _tspoint_gs;
    }
    if (strcmp(fcode,"BD130")==0 && strcmp(fcname,"dangerp")==0) {
      // dangerp.pft f_code BD130 Rock 
      return _dangerpoint_gs;
    }
    if (strcmp(fcode,"BD180")==0 && strcmp(fcname,"dangerp")==0) {
      // dangerp.pft f_code BD180 Wreck 
      return _dangerpoint_gs;
    }
    if (strcmp(fcode,"BE015")==0 && strcmp(fcname,"depthl")==0) {
      // depthl.lft f_code BE015 Depth Contour 
      return _dnline_gs;
    }
    if (strcmp(fcode,"BH000")==0 && strcmp(fcname,"aquecanl")==0) {
      // aquecanl.lft f_code BH000 Inland Water 
      return _dnline_gs;
    }
    if (strcmp(fcode,"BH000")==0 && strcmp(fcname,"aquecanl")==0) {
      // aquecanl.lft f_code BH000 Inland Water 
      return _dnline_gs;
    }
    if (strcmp(fcode,"BH000")==0 && strcmp(fcname,"inwatera")==0) {
      // inwatera.aft f_code BH000 Inland Water 
      return _dnarea_gs;
    }
    if (strcmp(fcode,"BH050")==0 && strcmp(fcname,"fishinda")==0) {
      // fishinda.aft f_code BH050 Fish Hatchery/Fish Farm/Marine Farm 
      return _dnarea_gs;
    }
    if (strcmp(fcode,"BH070")==0 && strcmp(fcname,"transtrc")==0) {
      // transtrc.pft f_code BH070 Ford 
      return _dnpoint_gs;
    }
    if (strcmp(fcode,"BH070")==0 && strcmp(fcname,"transtrl")==0) {
      // transtrl.lft f_code BH070 Ford 
      return _dnline_gs;
    }
    if (strcmp(fcode,"BH090")==0 && strcmp(fcname,"inwatera")==0) {
      // inwatera.aft f_code BH090 Land Subject to Inundation 
      return _poland_gs;
    }
    if (strcmp(fcode,"BH095")==0 && strcmp(fcname,"swampa")==0) {
      // swampa.aft f_code BH095 Marsh/Swamp 
      return _dnarea_gs;
    }
    if (strcmp(fcode,"BH120")==0 && strcmp(fcname,"miscp")==0) {
      // miscp.pft f_code BH120 Rapids 
      return _dnpoint_gs;
    }
    if (strcmp(fcode,"BH135")==0 && strcmp(fcname,"cropa")==0) {
      // cropa.aft f_code BH135 Rice Field 
      return _dnarea_gs;
    }
    if (strcmp(fcode,"BH140")==0 && strcmp(fcname,"watrcrsl")==0) {
      // watrcrsl.lft f_code BH140 River/Stream 
      return _dnline_gs;
    }
    if (strcmp(fcode,"BH155")==0 && strcmp(fcname,"extracta")==0) {
      // extracta.aft f_code BH155 Salt Evaporator 
      return _dnarea_gs;
    }
    if (strcmp(fcode,"BH155")==0 && strcmp(fcname,"extractp")==0) {
      // extractp.pft f_code BH155 Salt Evaporator 
      return _dnpoint_gs;
    }
    if (strcmp(fcode,"BH170")==0 && strcmp(fcname,"miscp")==0) {
      // miscp.pft f_code BH170 Spring/Water-Hole 
      return _dnpoint_gs;
    }
    if (strcmp(fcode,"BH180")==0 && strcmp(fcname,"miscp")==0) {
      // miscp.pft f_code BH180 Waterfall 
      return _dnline_gs;
    }
    if (strcmp(fcode,"BI020")==0 && strcmp(fcname,"miscl")==0) {
      // miscl.lft f_code BI020 Dam/Weir 
      return _dnline_gs;
    }
    if (strcmp(fcode,"BI020")==0 && strcmp(fcname,"miscp")==0) {
      // miscp.pft f_code BI020 Dam/Weir 
      return _dnpoint_gs;
    }
    if (strcmp(fcode,"BI030")==0 && strcmp(fcname,"miscp")==0) {
      // miscp.pft f_code BI030 Lock 
      return _dnpoint_gs;
    }
    if (strcmp(fcode,"BI040")==0 && strcmp(fcname,"miscp")==0) {
      // miscp.pft f_code BI040 Sluice Gate 
      return _dnpoint_gs;
    }
    if (strcmp(fcode,"BJ040")==0 && strcmp(fcname,"lndfrml")==0) {
      // lndfrml.lft f_code BJ040 Ice Cliff 

    }
    if (strcmp(fcode,"BJ065")==0 && strcmp(fcname,"seaicea")==0) {
      // seaicea.aft f_code BJ065 Ice Shelf 
	return _poice_gs;
    }
    if (strcmp(fcode,"BJ070")==0 && strcmp(fcname,"seaicea")==0) {
      // seaicea.aft f_code BJ070 Pack Ice 
	return _poice_gs;
    }
    if (strcmp(fcode,"BJ080")==0 && strcmp(fcname,"seaicea")==0) {
      // seaicea.aft f_code BJ080 Polar Ice 
	return _poice_gs;
    }
    if (strcmp(fcode,"BJ100")==0 && strcmp(fcname,"landicea")==0) {
      // landicea.aft f_code BJ100 Snow Field/Ice Field 
	return _poice_gs;
    }
    if (strcmp(fcode,"BJ110")==0 && strcmp(fcname,"tundraa")==0) {
      // tundraa.aft f_code BJ110 Tundra 
	return _vgtundr_gs;
    }
    if (strcmp(fcode,"CA010")==0 && strcmp(fcname,"contourl")==0) {
      // contourl.lft f_code CA010 Contour Line (Land) 
      return _hyline_gs;
    }
    if (strcmp(fcode,"CA030")==0 && strcmp(fcname,"elevp")==0) {
      // elevp.pft f_code CA030 Spot Elevation 
      return _elpoint_gs;
    }
    if (strcmp(fcode,"CA035")==0 && strcmp(fcname,"elevp")==0) {
      // elevp.pft f_code CA035 Inland Water Elevation 
      return _dnpoint_gs;
    }
    if (strcmp(fcode,"DA010")==0 && strcmp(fcname,"grounda")==0) {
      // grounda.aft f_code DA010 Ground Surface Element 
      return _poland_gs;
    }
    if (strcmp(fcode,"DB010")==0 && strcmp(fcname,"lndfrml")==0) {
      // lndfrml.lft f_code DB010 Bluff/Cliff/Escarpment 
      return _hyline_gs;
    }
    if (strcmp(fcode,"EA010")==0 && strcmp(fcname,"cropa")==0) {
      // cropa.aft f_code EA010 Cropland 
      return _vgrange_gs;
    }
    if (strcmp(fcode,"EB010")==0 && strcmp(fcname,"grassa")==0) {
      // grassa.aft f_code EB010 Grassland 
      return _vgrange_gs;
    }
    if (strcmp(fcode,"EB020")==0 && strcmp(fcname,"grassa")==0) {
      // grassa.aft f_code EB020 Scrub/Brush 
      return _vgrange_gs;
    }
    if (strcmp(fcode,"EC030")==0 && strcmp(fcname,"treesa")==0) {
      // treesa.aft f_code EC030 Trees 
      return _vgforst_gs;
    }
    if (strcmp(fcode,"FA000")==0 && strcmp(fcname,"libref")==0) {
      // libref.lft f_code FA000 Administrative Boundary 

    }
    if (strcmp(fcode,"FA000")==0 && strcmp(fcname,"polbndl")==0) {
      // polbndl.lft f_code FA000 Administrative Boundary 

    }
    if (strcmp(fcode,"FA001")==0 && strcmp(fcname,"polbnda")==0) {
      // polbnda.aft f_code FA001 Administrative Area 
      return _poland_gs;
    }
    if (strcmp(fcode,"FA001")==0 && strcmp(fcname,"polbndp")==0) {
      // polbndp.pft f_code FA001 Administrative Area 

    }
    if (strcmp(fcode,"FA020")==0 && strcmp(fcname,"polbndl")==0) {
      // polbndl.lft f_code FA020 Armistice Line 

    }
    if (strcmp(fcode,"FA030")==0 && strcmp(fcname,"polbndl")==0) {
      // polbndl.lft f_code FA030 Cease-Fire Line 

    }
    if (strcmp(fcode,"FA040")==0 && strcmp(fcname,"polbndl")==0) {
      // polbndl.lft f_code FA040 Claim Line 

    }
    if (strcmp(fcode,"FA060")==0 && strcmp(fcname,"polbndl")==0) {
      // polbndl.lft f_code FA060 Defacto Boundary 

    }
    if (strcmp(fcode,"FA070")==0 && strcmp(fcname,"polbnda")==0) {
      // polbnda.aft f_code FA070 Demilitarized Zone 
      return _poland_gs;
    }
    if (strcmp(fcode,"FA110")==0 && strcmp(fcname,"polbndl")==0) {
      // polbndl.lft f_code FA110 International Date Line 

    }
    if (strcmp(fcode,"FA170")==0 && strcmp(fcname,"polbnda")==0) {
      // polbnda.aft f_code FA170 Zone of Occupation 
      return _poland_gs;
    }
    if (strcmp(fcode,"GB005")==0 && strcmp(fcname,"aerofacp")==0) {
      // aerofacp.pft f_code GB005 Airport/Airfield 
      return _aepoint_gs;
    }
    if (strcmp(fcode,"SU001")==0 && strcmp(fcname,"mispopa")==0) {
      // mispopa.aft f_code SU001 Military Base 
      return _clarea_gs;
    }
    if (strcmp(fcode,"SU001")==0 && strcmp(fcname,"mispopp")==0) {
      // mispopp.pft f_code SU001 Military Base 
      return _pppoint_gs;
    }
    if (strcmp(fcode,"ZD040")==0 && strcmp(fcname,"bndtxt.tft")==0) {
      // bndtxt.tft f_code ZD040 Named Location 
      return _text_gs;
    }
    if (strcmp(fcode,"ZD040")==0 && strcmp(fcname,"hydrotxt.tft")==0) {
      // hydrotxt.tft f_code ZD040 Named Location 
      return _text_gs;
    }
    if (strcmp(fcode,"ZD040")==0 && strcmp(fcname,"indtxt.tft")==0) {
      // indtxt.tft f_code ZD040 Named Location 
      return _text_gs;
    }
    if (strcmp(fcode,"ZD040")==0 && strcmp(fcname,"libreft.tft")==0) {
      // libreft.tft f_code ZD040 Named Location 
      return _text_gs;
    }
    if (strcmp(fcode,"ZD040")==0 && strcmp(fcname,"phystxt.tft")==0) {
      // phystxt.tft f_code ZD040 Named Location 
      return _text_gs;
    }
    if (strcmp(fcode,"ZD040")==0 && strcmp(fcname,"poptxt.tft")==0) {
      // poptxt.tft f_code ZD040 Named Location 
      return _text_gs;
    }
    if (strcmp(fcode,"ZD040")==0 && strcmp(fcname,"transtxt.tft")==0) {
      // transtxt.tft f_code ZD040 Named Location 
      return _text_gs;
    }
    if (strcmp(fcode,"ZD040")==0 && strcmp(fcname,"utiltxt.tft")==0) {
      // utiltxt.tft f_code ZD040 Named Location 
      return _text_gs;
    }
    if (strcmp(fcode,"ZD045")==0 && strcmp(fcname,"bndtxt.tft")==0) {
      // bndtxt.tft f_code ZD045 Text Description 
      return _text_gs;
    }
    if (strcmp(fcode,"ZD045")==0 && strcmp(fcname,"hydrotxt.tft")==0) {
      // hydrotxt.tft f_code ZD045 Text Description 
      return _text_gs;
    }
    if (strcmp(fcode,"ZD045")==0 && strcmp(fcname,"indtxt.tft")==0) {
      // indtxt.tft f_code ZD045 Text Description 
      return _text_gs;
    }
    if (strcmp(fcode,"ZD045")==0 && strcmp(fcname,"libreft.tft")==0) {
      // libreft.tft f_code ZD045 Text Description 
      return _text_gs;
    }
    if (strcmp(fcode,"ZD045")==0 && strcmp(fcname,"phystxt.tft")==0) {
      // phystxt.tft f_code ZD045 Text Description 
      return _text_gs;
    }
    if (strcmp(fcode,"ZD045")==0 && strcmp(fcname,"poptxt.tft")==0) {
      // poptxt.tft f_code ZD045 Text Description 
      return _text_gs;
    }
    if (strcmp(fcode,"ZD045")==0 && strcmp(fcname,"transtxt.tft")==0) {
      // transtxt.tft f_code ZD045 Text Description 
      return _text_gs;
    }
    if (strcmp(fcode,"ZD045")==0 && strcmp(fcname,"utiltxt.tft")==0) {
      // utiltxt.tft f_code ZD045 Text Description 
      return _text_gs;
    }
    }


    // fall-through cases
    if (vf->type() == MapFeature::LineType) {
	return _unkline_gs;
    }
    else if (vf->type() == MapFeature::AreaType) {
	return _unkarea_gs;
    }
    else if (vf->type() == MapFeature::PointType)
	return _unkpoint_gs;
    else if (vf->type() == MapFeature::TextType)
      return _text_gs;
}
